
 -- This demo will show an example of how to easily
 -- create relatively deep virtual world with a
 -- relatively limited amount of code

 -- It displays a map of about 200 cities, and allows
 -- the arrow keys to switch between cities and
 -- city structures:

 --    UP: Next lower city (numerically)
 --  DOWN: Next higher city (numerically)
 --  LEFT: Next lower city structures (numerically)
 -- RIGHT: Next higher city structures (numerically)
 --  HOME: First city on the map
 --   END: Last city on the map
 --  PGUP: Skip 10 cities moving lower
 --  PGDN: Skip 10 cities moving higher

 --   ESC: Quit the program

 ----------

include get.e
include graphics.e
include igt.e

 ----------

constant ESC = 27

constant NUM_COLUMNS = 100, NUM_ROWS = 100

constant WARP_POINT = 1
constant SPECIAL_STRUCT_DATA = {"WarpPoint"}

constant VERY_LOW_PRICES = 1,
         LOW_PRICES = 2,
         AVG_PRICES = 3,
         HIGH_PRICES = 4,
         VERY_HIGH_PRICES = 5
constant PRICING_DATA = {"Very Low", "Low", "Average", "High", "Very High"}
constant PRICING_CHANGE = {0.25, 0.5, 1, 2, 4}

constant NO_ITEM = 1,
         CASH_ITEM = 2,
         SPECIAL_SWORD = 3,
         SPECIAL_SHIELD = 4,
         SPECIAL_ACCESSORY = 5,
         SPECIAL_SCROLL = 6,
         SPECIAL_ARMOR = 7,
         GEMSTONE_ITEM = 8
constant HIDDEN_ITEM_DATA = {"NONE", "Cash", "Ult. Sword", "Ult.Shield",
   "U.Accessy.", "Ult.Scroll", "Ult. Armor", "Gemstone"}

constant HOUSE_STRUCT = 1,
         SHOP_STRUCT = 2,
         TRAIN_HALL_STRUCT = 3,
         W_STORE_STRUCT = 4,
         SPECIAL_STRUCT = 5,
         INN_STRUCT = 6
constant STRUCT_DATA = {"House", "Shop", "Train.Hall", "Weap.Store",
   "Sp. Locale", "Inn"}

constant NUM_REG_SHOP_ITEMS = 3
constant REG_SHOP_ITEM_DATA =
   {{"Herb", 25}, {"Biscuit", 10}, {"Antidote", 100}}
constant POTION =         1,
         MEAT =           2,
         GEMSTONE =       3,
         POWER_GAUNTLET = 4,
         SPEED_BRACELET = 5,
         AURA_RING =      6,
         WARP_RING =      7,
         TITANS_BELT =    8
constant EXTRA_SHOP_ITEMS = {POTION, MEAT, GEMSTONE, POWER_GAUNTLET,
   SPEED_BRACELET, AURA_RING, WARP_RING, TITANS_BELT}
constant NUM_EXTRA_SHOP_ITEMS = length (EXTRA_SHOP_ITEMS)
constant EXTRA_SHOP_ITEM_DATA =
   {{"Potion", 200}, {"Meat", 100}, {"Gemstone", 100000},
    {"P.Gauntlet", 10000}, {"S.Bracelet", 1000}, {"Aura Ring", 5000},
    {"Warp Ring", 2000}, {"Titan Belt", 5000}}

constant NUM_REG_STORE_ITEMS = 3
constant REG_STORE_ITEM_DATA =
   {{"Wood Sword", 150}, {"WoodShield", 75}, {"Wood Armor", 100}}
constant SWORD_A  =  1,
         SWORD_B  =  2,
         SWORD_C  =  3,
         SWORD_X  =  4,
         SWORD_Y  =  5,
         SWORD_Z  =  6,
         SHIELD_A =  7,
         SHIELD_B =  8,
         SHIELD_C =  9,
         SHIELD_X = 10,
         SHIELD_Y = 11,
         SHIELD_Z = 12,
         ARMOR_A  = 13,
         ARMOR_B  = 14,
         ARMOR_C  = 15,
         ARMOR_X  = 16,
         ARMOR_Y  = 17,
         ARMOR_Z  = 18
constant EXTRA_STORE_ITEMS = {SWORD_A, SWORD_B, SWORD_C, SWORD_X, SWORD_Y,
   SWORD_Z, SHIELD_A, SHIELD_B, SHIELD_C, SHIELD_X, SHIELD_Y, SHIELD_Z,
   ARMOR_A, ARMOR_B, ARMOR_C, ARMOR_X, ARMOR_Y, ARMOR_Z}
constant NUM_EXTRA_STORE_ITEMS = length (EXTRA_STORE_ITEMS)
constant EXTRA_STORE_ITEM_DATA =
   {{"SteelSword", 250}, {"Silv.Sword", 400}, {"Fire Sword", 700},
    {"Myst.Sword", 1200}, {"Aura Sword", 2000}, {"PowerSword", 5000},
    {"St.Shield", 150}, {"Sil.Shield", 200}, {"FireShield", 400},
    {"Mys.Shield", 800}, {"AuraShield", 1200}, {"Pwr.Shield", 3000},
    {"SteelArmor", 200}, {"Silv.Armor", 300}, {"Fire Armor", 550},
    {"Myst.Armor", 1000}, {"Aura Armor", 1600}, {"PowerArmor", 4000}}


 ----------

constant price_set = new_igt_objectset ({{VERY_LOW_PRICES, 0.05},
                                         {LOW_PRICES, 0.20},
                                         {AVG_PRICES, 0.50},
                                         {HIGH_PRICES, 0.20},
                                         {VERY_HIGH_PRICES, 0.05}}),

special_locale_set = new_igt_objectset ({{WARP_POINT, 1.00}}),

hidden_item_set = new_igt_objectset ({{NO_ITEM, 0.85},
                                      {CASH_ITEM, 0.04},
                                      {SPECIAL_SWORD, 0.02},
                                      {SPECIAL_SHIELD, 0.02},
                                      {SPECIAL_ACCESSORY, 0.015},
                                      {SPECIAL_SCROLL, 0.02},
                                      {SPECIAL_ARMOR, 0.02},
                                      {GEMSTONE_ITEM, 0.015}}),

structure_set = new_igt_objectset ({{HOUSE_STRUCT, 0.45},
                                    {SHOP_STRUCT, 0.19},
                                    {TRAIN_HALL_STRUCT, 0.10},
                                    {W_STORE_STRUCT, 0.15},
                                    {SPECIAL_STRUCT, 0.06},
                                    {INN_STRUCT, 0.05}}),

city_name_set = new_igt_nameset ({"ARAGON", "THRENOPLE", "CADADOR",
                                  "MERIDIAN", "TELLOS", "ARANACIA",
                                  "CHEYRILL", "TANDOR", "VIVIDON",
                                  "WISPIN", "BELLIDUS", "EARLE",
                                  "GANDELA", "ISHTAR", "JULO", "LENAHON",
                                  "MIDWAVEN", "NEOPOLIS", "ODDELION",
                                  "PERICLES", "RASPANDA", "SHARD",
                                  "OPAL", "LAYME", "WASSILA", "YARK",
                                  "LANDIBAR", "DALIGUS", "ROSEHILL",
                                  "AVALON", "EDEN", "HORROWAY"})

 ----------

sequence name, extra_item_list
integer  base, pricing, num_ver_roads, num_hor_roads, lower, upper,
         num_npcs, population, num_structs, hidden_item, num_rooms,
         num_inhabitants, cash_amount, num_students, num_trainers,
         shop_size, store_size, num_extra_items, curr_extra,
         first_x, first_y, curr_x, curr_y, count, curr_clr,
         clr, L_id, has_city, k, still_continuing, curr_city,
         last_x, last_y, curr_struct, curr_num_structs
atom     prior_time, curr_time
object   junk, data_list

 ----------

type invalid_keypress (object x)

   if (not (integer (x))) then
      return 1
   end if

   if (x = ' ') or (x = '\n') or (x = ESC) or (x > 255) then
      return 0
   end if

   return 1

end type -- invalid_keypress()

 ----------

function max (object x, object y)

   return ((x > y) * x) + ((x <= y) * y)

end function -- max()

 ----------

function get_location_id (integer col, integer row)

   return ((col * NUM_COLUMNS) + row) * 7 + 1 -- * 7 avoids strong patterns

end function -- get_location_id()

 ----------

function is_city_at_location (integer locale_id)

   data_list = igt_values (locale_id, {{1, 49}}) -- 1 of 50 was to be a city,
                                                 -- but 49 isn't as patterned
   return (data_list[1] = 49)

end function -- is_city_at_location()

 ----------

function get_city_data (integer cid)

   data_list = igt_names (cid, {city_name_set},
                 {{NO_THREE_PLUS_LETTERS, NO_MULTIPLE_FIRST_LETTER}})
   name = data_list[1]

   data_list = igt_objects (cid, {price_set})
   pricing   = data_list[1]

   data_list = igt_values (cid, {{1,8}, {1,5}, {1,1000}, {1,100000}})

   num_ver_roads = data_list[1]
   num_hor_roads = data_list[2] + num_ver_roads - 3
   num_npcs = data_list[3]
   population = data_list[4]

   num_hor_roads = max (0, num_hor_roads)


    -- we want to make the boundaries for num_npcs and population
    -- be based on two other random values, so we take extra steps
   base = (num_ver_roads + 1) * (num_hor_roads + 1)
   lower = floor (base / 2)
   upper = base - 1
   num_npcs = remainder (num_npcs, (upper - lower + 1)) + lower

   lower = base * 10
   upper = base * 25
   population = remainder (population, (upper - lower + 1)) + lower

   num_structs = base

   return {name, pricing, num_ver_roads, num_hor_roads, num_npcs,
           population, num_structs}

end function -- get_city_data()

 ----------

function get_struct_id (integer cid, integer spot)

    -- try not to be too "neat" or we get subtle repeats (they
    -- seem to be based on Euphoria's random number generator)
   return (cid * 201) + spot

end function -- get_struct_id()

 ----------
 
function get_struct_at_spot (integer structid)

   data_list = igt_objects (structid, {structure_set})

   return data_list[1]

end function -- get_struct_at_spot()

 ----------

function get_house_data (integer structid)

   data_list = igt_objects (structid, {hidden_item_set})
   hidden_item = data_list[1]

   data_list = igt_values (structid, {{2,8}, {1,1000}, {-100,50}})

   base = data_list[1]
   num_inhabitants = data_list[2]
   cash_amount = data_list[3] * 100
   if (cash_amount < 0) then
      cash_amount = 0
   end if

   num_rooms = base

    -- we want the number of inhabitants to be dependant
    -- on the random base, so we make some adjustments
   lower = base - 2
   upper = base + 2
   num_inhabitants = remainder (num_inhabitants, (upper-lower+1)) + lower

   return {hidden_item, num_rooms, num_inhabitants, cash_amount}

end function -- get_house_data()

 ----------

function get_training_hall_data (integer structid)

   data_list = igt_values (structid, {{2,10}})
   num_students = data_list[1]
   num_trainers = rand (4)

   return {num_trainers, num_students}

end function -- get_training_hall_data()

 ----------

function get_inn_data (integer structid)

   data_list = igt_values (structid, {{2,5}})
   return data_list[1]

end function -- get_inn_data()

 ----------

function get_special_struct_data (integer structid)

   data_list = igt_objects (structid, {special_locale_set})
   return data_list[1]

end function -- get_special_struct_data()

 ----------

function get_shop_data (integer structid, integer prices)

   data_list = igt_values (structid,
                           {{prices*10, prices*20}, -- size in tiles
                           {2,5}})
   shop_size = data_list[1]
   num_extra_items = data_list[2]

   extra_item_list = repeat (0, num_extra_items)

    -- this works as long as we do immediately after calling
    -- 'igt_values', without resetting the random seed
   for i = 1 to num_extra_items do
      curr_extra = EXTRA_SHOP_ITEMS[rand(NUM_EXTRA_SHOP_ITEMS)]
      while (find (curr_extra, extra_item_list)) do
         curr_extra = EXTRA_SHOP_ITEMS[rand(NUM_EXTRA_SHOP_ITEMS)]
      end while
      extra_item_list[i] = curr_extra
   end for

   return {shop_size, extra_item_list}

end function -- get_shop_data()

 ----------

function get_store_data (integer structid, integer prices)

   data_list = igt_values (structid,
                           {{prices*10, prices*20}, -- size in tiles
                            {3,6}})
   store_size = data_list[1]
   num_extra_items = data_list[2]

   extra_item_list = repeat (0, num_extra_items)

    -- this works as long as we do immediately after calling
    -- 'igt_values', without resetting the random seed
   for i = 1 to num_extra_items do
      curr_extra = EXTRA_STORE_ITEMS[rand(NUM_EXTRA_STORE_ITEMS)]
      while (find (curr_extra, extra_item_list)) do
         curr_extra = EXTRA_STORE_ITEMS[rand(NUM_EXTRA_STORE_ITEMS)]
      end while
      extra_item_list[i] = curr_extra
   end for

   return {store_size, extra_item_list}

end function -- get_store_data()

 ----------

 -- returns the location of the first city,
 -- and the total number of cities; assumes
 -- screen is already in graphics mode 19
function draw_world_map ()

   integer t1, t2

   count = 0
   for r = 0 to NUM_ROWS - 1 do
   for c = 0 to NUM_COLUMNS - 1 do
      L_id = get_location_id (c, r)
      has_city = is_city_at_location (L_id)
      if (has_city) then
         t1 = c
         t2 = r
         if (count = 0) then
            first_x = c
            first_y = r
         end if
         clr = BROWN
         count = count + 1
      else
         clr = GREEN
      end if
      pixel (clr, {c,r})
   end for
   end for

   return {first_x, first_y, t1, t2, count}

end function -- draw_world_map()

 ----------

procedure draw_city (integer cid)

   sequence d
   atom     t

   d = get_city_data (cid)

    -- draw brown city area
   polygon (BROWN, 1, {{10,110}, {89,110}, {89,189}, {10,189}})

    -- draw the vertical roads
   if (d[3] = 1) then
      draw_line (WHITE, {{50,110}, {50,189}})
   else
      t = (80 / (d[3] + 1))
      for i = 1 to d[3] do
         draw_line (WHITE, {{i*t+10,110}, {i*t+10,189}})
      end for
   end if

    -- draw the horizonal roads
   if (d[4] > 0) then
      if (d[4] = 1) then
         draw_line (WHITE, {{10,150}, {89,150}})
      else
         t = (80 / (d[4] + 1))
         for i = 1 to d[4] do
            draw_line (WHITE, {{10,i*t+110}, {89,i*t+110}})
         end for
      end if
   end if

end procedure -- draw_city()

 ----------

procedure detail_city (integer cid)

   sequence d
   
   d = get_city_data (cid)

   for i = 1 to 6 do
      position (i,14)
      puts (1, "                          ") -- 26 spaces
   end for

   position (1,14)
   puts   (1, "NAME:   " & d[1])
   position (2,14)
   printf (1, "POP.:   %d", {d[6]})
   position (3,14)
   printf (1, "# NPCS: %d", {d[5]})
   position (4,14)
   puts   (1, "PRICES: " & PRICING_DATA[d[2]])
   position (6,14)
   printf (1, "# STRUCTURES: %d", {d[7]})

end procedure -- detail_city()

 ----------

procedure detail_struct (integer cid, integer spot)

   sequence d, extra_shop_items, extra_store_items
   integer  sid, s
   atom     pricing

   d = get_city_data (cid)
   pricing = PRICING_CHANGE[d[2]]

   sid = get_struct_id (cid, spot)

   for i = 10 to 24 do
      position (i,14)
      puts (1, "                          ") -- 26 spaces
   end for

   position (10,14)
   printf (1, "   STRUCTURE # %d:", {spot})

   position (12,14)
   s = get_struct_at_spot (sid)
   puts   (1, "TYPE: " & STRUCT_DATA[s])
   position (13,14)

   if (s = INN_STRUCT) then
      printf (1, "NUM. ROOMS: %d", {get_inn_data (sid)})

   elsif (s = HOUSE_STRUCT) then
      d = get_house_data (sid)
      printf (1, "NUM. ROOMS: %d", {d[2]})
      position (14,14)
      printf (1, "NUM. INHABITANTS: %d", {d[3]})
      position (15,14)
      printf (1, "CASH HIDDEN INSIDE: $%d", {d[4]})
      position (16,14)
      puts   (1, "HIDDEN ITEM: " & HIDDEN_ITEM_DATA[d[1]])

   elsif (s = SHOP_STRUCT) then
      d = get_shop_data (sid, d[2])
      printf (1, "SIZE: %d tiles", {d[1]})
      position (14,14)
      puts (1, "ITEMS FOR SALE:")
      extra_shop_items = d[2]
      for i = 1 to NUM_REG_SHOP_ITEMS do
         position (14+i,14)
         printf (1, "  %-10s ($%d)", {REG_SHOP_ITEM_DATA[i][1],
                               floor (REG_SHOP_ITEM_DATA[i][2] * pricing)})
      end for
      for i = 1 to length (extra_shop_items) do
         position (14+NUM_REG_SHOP_ITEMS+i,14)
         printf (1, "  %-10s ($%d)", {EXTRA_SHOP_ITEM_DATA[extra_shop_items[i]][1],
                               floor (EXTRA_SHOP_ITEM_DATA[extra_shop_items[i]][2] * pricing)})
      end for

   elsif (s = TRAIN_HALL_STRUCT) then
      d = get_training_hall_data (sid)
      printf (1, "NUM. TRAINERS: %d", {d[1]})
      position (14,14)
      printf (1, "NUM. STUDENTS: %d", {d[2]})

   elsif (s = W_STORE_STRUCT) then
      d = get_store_data (sid, d[2])
      printf (1, "SIZE: %d tiles", {d[1]})
      position (14,14)
      puts (1, "ITEMS FOR SALE:")
      extra_store_items = d[2]
      for i = 1 to NUM_REG_STORE_ITEMS do
         position (14+i,14)
         printf (1, "  %-10s ($%d)", {REG_STORE_ITEM_DATA[i][1],
                               floor (REG_STORE_ITEM_DATA[i][2] * pricing)})
      end for
      for i = 1 to length (extra_store_items) do
         position (14+NUM_REG_STORE_ITEMS+i,14)
         printf (1, "  %-10s ($%d)", {EXTRA_STORE_ITEM_DATA[extra_store_items[i]][1],
                               floor (EXTRA_STORE_ITEM_DATA[extra_store_items[i]][2] * pricing)})
      end for

   elsif (s = SPECIAL_STRUCT) then
      puts (1, "SPECIFICS: " & SPECIAL_STRUCT_DATA
            [get_special_struct_data(sid)])

   else
      -- error; ignore, no data
   end if

end procedure -- detail_struct()

 ----------

procedure perform_action (integer a)

   object temp

   if (a = '8') or (a = 328) then
       -- go to prior city
      if (curr_x != first_x) or (curr_y != first_y) then
         curr_x = curr_x - 1
         if (curr_x < 0) then
            curr_x = NUM_COLUMNS - 1
            curr_y = curr_y - 1
         end if
         temp = get_location_id (curr_x, curr_y)
         while (not (is_city_at_location (temp))) do
            curr_x = curr_x - 1
            if (curr_x < 0) then
               curr_x = NUM_COLUMNS - 1
               curr_y = curr_y - 1
            end if
            temp = get_location_id (curr_x, curr_y)
         end while
         curr_city = curr_city - 1
         curr_struct = 1
         draw_city   (temp)
         detail_city (temp)
         detail_struct (temp, curr_struct)
         temp = get_city_data (temp)
         curr_num_structs = temp[7]
      end if

   elsif (a = '2') or (a = 336) then
       -- go to next city
      if (curr_city != count) then
         curr_x = curr_x + 1
         if (curr_x > NUM_COLUMNS - 1) then
            curr_x = 0
            curr_y = curr_y + 1
         end if
         temp = get_location_id (curr_x, curr_y)
         while (not (is_city_at_location (temp))) do
            curr_x = curr_x + 1
            if (curr_x > NUM_COLUMNS - 1) then
               curr_x = 0
               curr_y = curr_y + 1
            end if
            temp = get_location_id (curr_x, curr_y)
         end while
         curr_city = curr_city + 1
         curr_struct = 1
         draw_city   (temp)
         detail_city (temp)
         detail_struct (temp, curr_struct)
         temp = get_city_data (temp)
         curr_num_structs = temp[7]
      end if

   elsif (a = '4') or (a = 331) then
      if (curr_struct > 1) then
         curr_struct = curr_struct - 1
      end if
      detail_struct (get_location_id (curr_x, curr_y), curr_struct)

   elsif (a = '6') or (a = 333) then
      if (curr_struct < curr_num_structs) then
         curr_struct = curr_struct + 1
      end if
      detail_struct (get_location_id (curr_x, curr_y), curr_struct)
   
   elsif (a = 329) then -- PGUP; move up 10 cities
      for i = 1 to 10 do
         perform_action ('8')         
      end for
      temp = get_location_id (curr_x, curr_y)
      draw_city   (temp)
      detail_city (temp)
      detail_struct (temp, curr_struct)
      temp = get_city_data (temp)
      curr_num_structs = temp[7]

   elsif (a = 337) then -- PGDN; move down 10 cities
      for i = 1 to 10 do
         perform_action ('2')
      end for
      temp = get_location_id (curr_x, curr_y)
      draw_city   (temp)
      detail_city (temp)
      detail_struct (temp, curr_struct)
      temp = get_city_data (temp)
      curr_num_structs = temp[7]

   elsif (a = 327) then -- HOME; move to first city
      curr_city = 1
      curr_x = first_x
      curr_y = first_y
      curr_struct = 1
      temp = get_location_id (curr_x, curr_y)
      draw_city   (temp)
      detail_city (temp)
      detail_struct (temp, curr_struct)
      temp = get_city_data (temp)
      curr_num_structs = temp[7]

   elsif (a = 335) then -- END; move to last city
      curr_city = count
      curr_x = last_x
      curr_y = last_y
      curr_struct = 1
      temp = get_location_id (curr_x, curr_y)
      draw_city   (temp)
      detail_city (temp)
      detail_struct (temp, curr_struct)
      temp = get_city_data (temp)
      curr_num_structs = temp[7]

   elsif (a = ESC) or (a = ' ') or (a = '\n') or (a > 255) then
      still_continuing = 0

   end if

end procedure -- perform_action()

 ----------

procedure main ()

    -- go to the proper graphics mode
   junk = graphics_mode (19)

    -- draw the map, and record the location of the first city
   junk = draw_world_map ()
   curr_x = junk[1]
   curr_y = junk[2]
   last_x = junk[3]
   last_y = junk[4]
   curr_city = 1
   curr_struct = 1

    -- draw the city & info
   L_id = get_location_id (curr_x, curr_y)
   draw_city   (L_id)
   detail_city (L_id)
   detail_struct (L_id, curr_struct)
   junk = get_city_data (L_id)
   curr_num_structs = junk[7]

    -- continue to blink the current city twice a second
    -- until user does something
   still_continuing = 1
   while (still_continuing) do

      curr_clr = BROWN
      prior_time = time ()
      k = get_key ()
      while (invalid_keypress (k)) do
         curr_time = time ()
         if (prior_time + 0.5 <= curr_time) then
            prior_time = curr_time
            curr_clr = YELLOW - curr_clr + BROWN
            pixel (curr_clr, {curr_x, curr_y})
         end if
         k = get_key ()
      end while
   
      pixel (BROWN, {curr_x, curr_y})

      perform_action (k)

   end while

    -- return to normal graphics mode
   junk = graphics_mode (3)

end procedure -- main()

 ----------

main ()

